/*
 * Project: VizKit
 * Version: 3.3.7
 
 * Date: 20150503
 * File: VisualAudioData.h
 *
 */

/***************************************************************************

Copyright (c) 2004-2015 Heiko Wichmann (http://www.imagomat.de/vizkit)


This software is provided 'as-is', without any expressed or implied warranty. 
In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; 
   you must not claim that you wrote the original software. 
   If you use this software in a product, an acknowledgment 
   in the product documentation would be appreciated 
   but is not required.

2. Altered source versions must be plainly marked as such, 
   and must not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source distribution.

 ***************************************************************************/

#ifndef VisualAudioData_h
#define VisualAudioData_h

#include "VisualTypes.h"


namespace VizKit {

	/**
	 * Raw audio data as received with pulse message via iTunes visualizer interface. 
	 */
	class VisualAudioData {
	
	public:

		/**
		 * Constructor.
		 */
		VisualAudioData();

		/**
		 * Destructor.
		 */
		~VisualAudioData();

		/**
		 * Copy constructor.
		 * @param other Another VisualAudioData.
		 */
		VisualAudioData(const VisualAudioData& other);

		/**
		 * Assignment operator.
		 */			
		VisualAudioData& operator=(const VisualAudioData& other);
		
		/**
		 * Equality operator.
		 * @param other Another VisualAudioData.
		 */
		bool operator==(const VisualAudioData& other);

		/**
		 * Inequality operator.
		 * @param other Another VisualAudioData.
		 */
		bool operator!=(const VisualAudioData& other);
        
        /**
		 * Sets the number of waveform channels.
		 * @param aWaveformChannelCount The number of waveform channels.
		 */
		void setNumberOfWaveformChannels(uint8 aWaveformChannelCount);
        
		/**
		 * Sets the number of waveform entries.
		 * @param aWaveformEntryCount The number of waveform entries.
		 */
		void setNumberOfWaveformEntries(uint16 aWaveformEntryCount);

		/**
		 * Sets the waveform data.
		 * @param someWaveformData The waveform data.
		 */		
		void setWaveformData(const uint8* const someWaveformData);
        
        /**
		 * Sets the number of spectrum channels.
		 * @param aSpectrumChannelCount The number of spectrum channels.
		 */
		void setNumberOfSpectrumChannels(uint8 aSpectrumChannelCount);
        
		/**
		 * Sets the number of spectrum entries.
		 * @param aSpectrumEntryCount The number of spectrum entries.
		 */
		void setNumberOfSpectrumEntries(uint16 aSpectrumEntryCount);
        
		/**
		 * Sets the spectrum data.
		 * @param someSpectrumData The spectrum data.
		 */		
		void setSpectrumData(const uint8* const someSpectrumData);

		/**
		 * Returns the number of waveform channels.
		 * @return The number of waveform channels.
		 */
		uint8 getNumberOfWaveformChannels(void) const;
        
        /**
		 * Returns the number of waveform entries.
		 * @return The number of waveform entries.
		 */
		uint16 getNumberOfWaveformEntries(void) const;
        
        /**
		 * Returns the waveform data.
		 * @return The waveform data.
		 */
		const uint8* const getWaveformData(void) const;
        
        /**
		 * Returns the number of spectrum channels.
		 * @return The number of spectrum channels.
		 */
		uint8 getNumberOfSpectrumChannels(void) const;
        
        /**
		 * Returns the number of spectrum entries.
		 * @return The number of spectrum entries.
		 */
		uint16 getNumberOfSpectrumEntries(void) const;
        
        /**
		 * Returns the spectrum data.
		 * @return The spectrum data.
		 */
		const uint8* const getSpectrumData(void) const;


	private:
        
        /** The number of waveform channels. */
        uint8 waveformChannelCount;
        
        /** The number of waveform entries. */
        uint16 waveformEntryCount;
        
        /** The waveform data. */
        uint8* waveformData;
        
        /** The number of spectrum channels. */
        uint8 spectrumChannelCount;
        
        /** The number of spectrum entries. */
        uint16 spectrumEntryCount;
        
        /** The spectrum data. */
        uint8* spectrumData;
			
		/**
		 * Copy method for assignment operator and copy constructor.
		 * @param other Another VisualAudioData.
		 */
		void copy(const VisualAudioData& other);

	};
	
}

#endif /* VisualAudioData_h */
