/*
 * Project: VizKit
 * Version: 3.3.7
 
 * Date: 20150503
 * File: VisualGraphicsCore.cpp
 *
 */

/***************************************************************************

Copyright (c) 2004-2015 Heiko Wichmann (http://www.imagomat.de/vizkit)


This software is provided 'as-is', without any expressed or implied warranty. 
In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; 
   you must not claim that you wrote the original software. 
   If you use this software in a product, an acknowledgment 
   in the product documentation would be appreciated 
   but is not required.

2. Altered source versions must be plainly marked as such, 
   and must not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source distribution.

 ***************************************************************************/

#include "VisualGraphicsCore.h"
#include "VisualErrorHandling.h"
#include "VisualConfiguration.h"

#if TARGET_OS_MAC
// file type of VisualGraphicsCore.cpp is set to "Objective-C++ source"
#import <Cocoa/Cocoa.h>
#include "VisualCustomization.h"
#import "VisualView.h"
#endif


#if TARGET_OS_WIN
#include <stdio.h> // sprintf
#endif


#if TARGET_OS_WIN
#include <GL/gl.h>
#include <GL/glu.h>
#endif


using namespace VizKit;


VisualGraphicsCore* VisualGraphicsCore::theVisualGraphicsCore = NULL;  // singleton


VisualGraphicsCore::VisualGraphicsCore() 
{
#if TARGET_OS_WIN
    ctx = NULL;
	windowDC = NULL;
#endif
}


VisualGraphicsCore::~VisualGraphicsCore() 
{
	// null
}


VisualGraphicsCore* VisualGraphicsCore::getInstance() 
{
    if (theVisualGraphicsCore == NULL) 
    {
		theVisualGraphicsCore = new VisualGraphicsCore;
    }
	return theVisualGraphicsCore;
}


void VisualGraphicsCore::dispose() 
{
    if (theVisualGraphicsCore != NULL) 
    {
        delete theVisualGraphicsCore;
		theVisualGraphicsCore = NULL;
    }
}


bool VisualGraphicsCore::setupContext() 
{

	bool success = true;
	char errLog[512];
	
#if TARGET_OS_MAC
    
    NSRect aRect = NSMakeRect(0.0, 0.0, canvasRect.width, canvasRect.height);
    visualView = [[VisualView alloc] initWithFrame:aRect];
	if (visualView != nil)
	{
		[visualView setAutoresizingMask: (NSViewWidthSizable | NSViewHeightSizable)];
		[platformView addSubview:visualView];
	}
	else
	{
		sprintf(errLog, "error: unable to setup visualView in file: %s (line: %d) [%s])", __FILE__, __LINE__, __FUNCTION__);
		writeLog(errLog);
        success = false;
	}
    
    [visualView setup];
    
#endif

#if TARGET_OS_WIN
	uint8 errNum = theVisualGraphicsCore->setupWGL();
	if (errNum != 0) 
    {
		sprintf(errLog, "error %d on setupWGL in file: %s (line: %d) [%s])", errNum, __FILE__, __LINE__, __FUNCTION__);
		writeLog(errLog);
		success = false;
	}
#endif
    
	return success;
}

#if TARGET_OS_MAC
void VisualGraphicsCore::startRendering()
{
    [visualView start];
}
#endif

void VisualGraphicsCore::cleanupContext()
{
#if TARGET_OS_MAC
    [visualView cleanup];
    [visualView removeFromSuperview];
    [visualView release];
    visualView = nil;
#endif
#if TARGET_OS_WIN
    this->cleanUpWGL();
#endif
}


uint16 VisualGraphicsCore::setCurrentContext() 
{
#if TARGET_OS_WIN
	if (!wglMakeCurrent(theVisualGraphicsCore->windowDC, theVisualGraphicsCore->ctx)) {
		writeLastErrorLog("setCurrentContext");
        return 1;
	}
#endif
	return 0;
}


#if TARGET_OS_WIN

uint8 VisualGraphicsCore::setupWGL() 
{

    PIXELFORMATDESCRIPTOR pfd;
    memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));

    pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 24;
	pfd.cRedBits = 8;
	pfd.cGreenBits = 8;
	pfd.cBlueBits = 8;
	pfd.cAlphaBits = 8;
    pfd.cDepthBits = 16;

	pfd.iLayerType = PFD_MAIN_PLANE;

	int aPixelFormat = ChoosePixelFormat(theVisualGraphicsCore->windowDC, &pfd);
    if (aPixelFormat == 0) 
	{
		writeLastErrorLog("err: setupWGL1");
        return 1;
	}
	
	if (SetPixelFormat(theVisualGraphicsCore->windowDC, aPixelFormat, &pfd) == FALSE) 
	{
		writeLastErrorLog("err: setupWGL2");
        return 2;
	}

	theVisualGraphicsCore->ctx = wglCreateContext(theVisualGraphicsCore->windowDC);
    if (theVisualGraphicsCore->ctx == NULL) 
	{
		writeLastErrorLog("err: setupWGL3");
        return 3;
	}

	if (wglMakeCurrent(theVisualGraphicsCore->windowDC, theVisualGraphicsCore->ctx) == FALSE) 
	{
		writeLastErrorLog("err: setupWGL4");
        return 4;
	}

    // enable v-sync if WGL_EXT_swap_control is supported
	/*
    PFNWGLSWAPINTERVALEXTPROC wglSwapIntervalEXT = NULL;
    wglSwapIntervalEXT = (PFNWGLSWAPINTERVALEXTPROC)wglGetProcAddress("wglSwapIntervalEXT");
    if (wglSwapIntervalEXT != NULL)
        wglSwapIntervalEXT(1);
	*/

	return 0;
}


void VisualGraphicsCore::cleanUpWGL() 
{

    wglMakeCurrent(NULL, NULL);

    if (theVisualGraphicsCore->ctx != NULL) {
        wglDeleteContext(theVisualGraphicsCore->ctx);
        theVisualGraphicsCore->ctx = NULL;
    }

    if (theVisualGraphicsCore->windowDC) {
		ReleaseDC(theVisualGraphicsCore->platformView, theVisualGraphicsCore->windowDC);
        theVisualGraphicsCore->windowDC = NULL;
    }
}


HDC VisualGraphicsCore::getWindowDC() 
{
	if (theVisualGraphicsCore->windowDC) 
    {
		return theVisualGraphicsCore->windowDC;
	} 
    else 
    {
		// if the HDC has not been set yet for the visualizer, we return the HDC of the main screen
		return GetDC(NULL);
	}
}

#endif


void VisualGraphicsCore::setPlatformView(VISUAL_PLATFORM_VIEW aView)
{
    theVisualGraphicsCore->platformView = aView;
#if TARGET_OS_WIN
    theVisualGraphicsCore->windowDC = GetDC(aView);
#endif
    this->refreshCanvasRect();
}


const VISUAL_PLATFORM_VIEW VisualGraphicsCore::getPlatformView() 
{
    return theVisualGraphicsCore->platformView;
}


void VisualGraphicsCore::refreshCanvasRect()
{
#if TARGET_OS_MAC
    NSRect aBoundsRect;
    if ([theVisualGraphicsCore->platformView respondsToSelector:@selector(convertRectToBacking:)])
    {
        aBoundsRect = [theVisualGraphicsCore->platformView convertRectToBacking:[theVisualGraphicsCore->platformView bounds]];
    }
    else
    {
        aBoundsRect = [theVisualGraphicsCore->platformView bounds];
    }
    canvasRect.width = aBoundsRect.size.width;
    canvasRect.height = aBoundsRect.size.height;
#endif
#if TARGET_OS_WIN
	RECT aRect;
	BOOL success = GetClientRect(theVisualGraphicsCore->platformView, &aRect);
	canvasRect.width = aRect.right - aRect.left;
	canvasRect.height = aRect.bottom - aRect.top;
#endif
}


PixelRect VisualGraphicsCore::getCanvasRect() 
{
	return this->canvasRect;
}


float VisualGraphicsCore::getScreenScaleFactor()
{
    float aScale = 1.0;
#if TARGET_OS_MAC
    if ([[this->platformView window] respondsToSelector:@selector(backingScaleFactor)])
    {
        CGFloat aMacScale = [[this->platformView window] backingScaleFactor];
        aScale = (float)aMacScale;
    }
#endif
    return aScale;
}


void VisualGraphicsCore::finishGLDrawing() 
{
    //glFinish();
    //glFlush();
#if TARGET_OS_WIN
	SwapBuffers(theVisualGraphicsCore->windowDC);
#endif
}

