/*
 * Project: VizKit
 * Version: 3.3.7
 
 * Date: 20150503
 * File: VisualTextAutoScrollView.cpp
 *
 */

/***************************************************************************
 
 Copyright (c) 2004-2015 Heiko Wichmann (http://www.imagomat.de/vizkit)
 
 
 This software is provided 'as-is', without any expressed or implied warranty. 
 In no event will the authors be held liable for any damages
 arising from the use of this software.
 
 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it
 freely, subject to the following restrictions:
 
 1. The origin of this software must not be misrepresented; 
 you must not claim that you wrote the original software. 
 If you use this software in a product, an acknowledgment 
 in the product documentation would be appreciated 
 but is not required.
 
 2. Altered source versions must be plainly marked as such, 
 and must not be misrepresented as being the original software.
 
 3. This notice may not be removed or altered from any source distribution.
 
 ***************************************************************************/

#include "VisualTextAutoScrollView.h"
#include "VisualStyledString.h"
#include "VisualGraphics.h"
#include "VisualPreferences.h"
#include "VisualDataStore.h"
#include "VisualImage.h"
#include "VisualCamera.h"
#include "VisualVertex.h"
#include "VisualErrorHandling.h"
#include "VisualPlayerState.h"

using namespace VizKit;

#include <math.h>
#ifndef M_PI
#define M_PI 3.14159265358979323846264338327950288
#endif



VisualTextAutoScrollView::VisualTextAutoScrollView() 
{
    textScrollViewStyledString = NULL;
    firstTextLineIndexToShow = 0;
    lastTextLineIndexToShow = 0;
    longestTextLinePixelWidth = 0;
    VisualGraphics* theVisualGraphics = VisualGraphics::getInstance();
    displayedPixelHeight = theVisualGraphics->getCanvasPixelHeight();
}


VisualTextAutoScrollView::~VisualTextAutoScrollView() 
{
    if (textScrollViewStyledString != NULL)
    {
        delete textScrollViewStyledString;
    }
    std::vector<VisualTextAutoScrollViewTextLineImageRecord *>::iterator it;
    for (it = textScrollViewImageVector.begin(); it != textScrollViewImageVector.end(); it++)
    {
		delete *it;
	}
    textScrollViewImageVector.clear();
}


VisualTextAutoScrollView::VisualTextAutoScrollView(const VisualTextAutoScrollView& other) : VisualObject(other) 
{
	copy(other);
}


VisualTextAutoScrollView& VisualTextAutoScrollView::operator=(const VisualTextAutoScrollView& other) 
{
	if (this == &other) return *this;
	
	VisualObject::operator=(other);
	
	this->copy(other);
	
	return *this;
}


bool VisualTextAutoScrollView::operator<(const VisualTextAutoScrollView& other) const 
{
	return (this->VisualObject::getIdentifier() < other.VisualObject::getIdentifier());
}


bool VisualTextAutoScrollView::operator==(const VisualTextAutoScrollView& other) const 
{
	return (this->VisualObject::getIdentifier() == other.VisualObject::getIdentifier());
}


bool VisualTextAutoScrollView::operator!=(const VisualTextAutoScrollView& other) const 
{
	return !(*this == other);
}


VisualTextAutoScrollView* VisualTextAutoScrollView::clone(void) const 
{
	return new VisualTextAutoScrollView(*this);
}


void VisualTextAutoScrollView::copy(const VisualTextAutoScrollView& other) 
{
    if (other.textScrollViewStyledString != NULL)
    {
		this->textScrollViewStyledString = new VisualStyledString(*(other.textScrollViewStyledString));
	}
    else
    {
		this->textScrollViewStyledString = NULL;
	}
}


void VisualTextAutoScrollView::setStyledString(VisualStyledString& aStyledString)
{
    if (this->textScrollViewStyledString != NULL)
    {
        delete this->textScrollViewStyledString;
    }
    this->textScrollViewStyledString = aStyledString.clone();
    this->refreshLayout();
}


void VisualTextAutoScrollView::refreshTextlineImageData(float aProgressValue)
{
    this->calculateFirstAndLastTextLineIndexToShow(aProgressValue);

    uint32 aNumberOfLines = this->textScrollViewStyledString->getNumberOfLines();
	VisualStringStyle aStringStyle = this->textScrollViewStyledString->getStyle();
    // font color of the style is set later when drawing
    // (not when creating the bitmap)
    aStringStyle.fontColor.r = 1.0;
	aStringStyle.fontColor.g = 1.0;
	aStringStyle.fontColor.b = 1.0;
    aStringStyle.fontColor.a = 1.0;
	aStringStyle.fontColor.red = 1.0;
	aStringStyle.fontColor.green = 1.0;
	aStringStyle.fontColor.blue = 1.0;
    aStringStyle.fontColor.alpha = 1.0;

	for (size_t i = 0; i < aNumberOfLines; i++)
    {
        bool needsToShow = ((i >= this->firstTextLineIndexToShow) && (i <= this->lastTextLineIndexToShow));
        
        std::vector<VisualTextAutoScrollViewTextLineImageRecord *>::iterator it = this->textScrollViewImageVector.begin() + i;
        VisualImage *anOldTextLineImage = (*it)->textImage;
        if (anOldTextLineImage->isEmpty() && (needsToShow == true))
        {
            VisualString* aSingleLineString = this->textScrollViewStyledString->createStringWithLine(i);
            VisualStyledString aStyledString;
            aStyledString.initWithVisualString(*aSingleLineString);
            aStyledString.setStyle(aStringStyle);
            VisualImage *aNewTextLineImage = VisualImage::createWithStyledString(aStyledString);
			if (aNewTextLineImage != NULL)
			{
				delete anOldTextLineImage;
				(*it)->textImage = aNewTextLineImage;
			}
			delete aSingleLineString;
        }
        else if ((anOldTextLineImage->isEmpty() == false) && (needsToShow == false))
        {
            VisualImage *aNewEmptyTextLineImage = new VisualImage;
            aNewEmptyTextLineImage->setBlendMode(kReplace);
            delete anOldTextLineImage;
            (*it)->textImage = aNewEmptyTextLineImage;
        }
	}
}


void VisualTextAutoScrollView::calculateFirstAndLastTextLineIndexToShow(float aProgressValue)
{
    uint32 aNumberOfLines = this->textScrollViewStyledString->getNumberOfLines();
    
    this->firstTextLineIndexToShow = 0;
    this->lastTextLineIndexToShow = (aNumberOfLines - 1);
    
    if (this->completeTextLinesPixelHeight() > this->displayedPixelHeight)
    {
        // we can not display complete amount of text lines
        
        uint32 aPixelHeightDelta = this->completeTextLinesPixelHeight() - this->displayedPixelHeight;
        uint32 anOffset = (uint32)((float)aPixelHeightDelta * aProgressValue);
        
        size_t aFirstTextLineToShowHeight = 0;
        
        bool searchForAdditionalLines = true;
        for (uint32 i = 0; i < aNumberOfLines; i++)
        {
            std::vector<VisualTextAutoScrollViewTextLineImageRecord *>::iterator it = this->textScrollViewImageVector.begin() + i;
            if ((*it)->textImageVerticalPixelPosition + (*it)->textImagePixelRect.height > anOffset)
            {
                aFirstTextLineToShowHeight = (*it)->textImagePixelRect.height;
                this->firstTextLineIndexToShow = i;
                if (i == (aNumberOfLines - 1))
                {
                    // last line
                    this->lastTextLineIndexToShow = i;
                    searchForAdditionalLines = false;
                }
                else if (aFirstTextLineToShowHeight > this->displayedPixelHeight)
                {
                    // first line covers displayed pixel height
                    this->lastTextLineIndexToShow = i;
                    searchForAdditionalLines = false;
                }
                break;
            }
        }
        
        if (searchForAdditionalLines == true)
        {
            size_t aStringBitmapHeight = aFirstTextLineToShowHeight;
            
            for (uint32 i = (this->firstTextLineIndexToShow + 1); i < aNumberOfLines; i++)
            {
                std::vector<VisualTextAutoScrollViewTextLineImageRecord *>::iterator it = this->textScrollViewImageVector.begin() + i;
                size_t aPixelRectHeight = (*it)->textImagePixelRect.height;
                aStringBitmapHeight += aPixelRectHeight;
                if ((aStringBitmapHeight - aFirstTextLineToShowHeight) >= this->displayedPixelHeight)
                {
                    this->lastTextLineIndexToShow = i;
                    break;
                }
            }
            // catch in case no line found
            if (aStringBitmapHeight < this->displayedPixelHeight)
            {
                this->lastTextLineIndexToShow = (aNumberOfLines - 1);
            }
        }
    }
}


const VisualStyledString* const VisualTextAutoScrollView::getStyledString()
{
    return this->textScrollViewStyledString;
}


void VisualTextAutoScrollView::setStringStyle(VisualStringStyle aStringStyle)
{
    this->textScrollViewStyledString->setStyle(aStringStyle);
}


VisualStringStyle VisualTextAutoScrollView::getStringStyle()
{
    VisualStringStyle& aStringStyle = this->textScrollViewStyledString->getStyle();
    return aStringStyle;
}


void VisualTextAutoScrollView::refreshLayout()
{
    
    // cleanup
    std::vector<VisualTextAutoScrollViewTextLineImageRecord *>::iterator it;
    for (it = this->textScrollViewImageVector.begin(); it != this->textScrollViewImageVector.end(); it++)
    {
        VisualTextAutoScrollViewTextLineImageRecord *aTextLineRecord = *it;
        delete aTextLineRecord->textImage;
		delete *it;
	}
    this->textScrollViewImageVector.clear();
    this->longestTextLinePixelWidth = 0;
    
    VisualGraphics *theVisualGraphics = VisualGraphics::getInstance();
    size_t aMaxPixelWidth = theVisualGraphics->getCanvasPixelWidth();
	VisualStringStyle& aStringStyle = this->textScrollViewStyledString->getStyle();
    
    size_t anAccumulatedTextLinesPixelHeight = 0;
    uint32 numberOfLines = this->textScrollViewStyledString->getNumberOfLines();
    for (uint32 i = 0; i < numberOfLines; i++)
    {
        VisualTextAutoScrollViewTextLineImageRecord *aTextLineRecord = new VisualTextAutoScrollViewTextLineImageRecord;
        VisualImage *anEmptyImage = new VisualImage;
        aTextLineRecord->textImage = anEmptyImage;
        VisualString* aSingleLineString = this->textScrollViewStyledString->createStringWithLine(i);
        if (aSingleLineString->isEmpty())
        {
            // blank line
            delete aSingleLineString;
            // create string with sample content to evaluate line height
            aSingleLineString = VisualString::createWithUTF8String("Tg");
        }
        VisualStyledString aSingleLineStyledString;
        aSingleLineStyledString.initWithVisualString(*aSingleLineString);
        aSingleLineStyledString.setStyle(aStringStyle);
        PixelRect aPixelRect = aSingleLineStyledString.bitmapPixelRect(aMaxPixelWidth, 0);
        aTextLineRecord->textImagePixelRect = aPixelRect;
        aTextLineRecord->textImageVerticalPixelPosition = anAccumulatedTextLinesPixelHeight;
        anAccumulatedTextLinesPixelHeight += aPixelRect.height;
        if (aPixelRect.width > this->longestTextLinePixelWidth)
        {
            this->longestTextLinePixelWidth = aPixelRect.width;
        }
        textScrollViewImageVector.insert(textScrollViewImageVector.begin() + i, aTextLineRecord);
        delete aSingleLineString;
    }
}


void VisualTextAutoScrollView::draw(float aCurrAudioPlayProgressValue)
{
    
    float aProgressValue = this->convertProgressValueToInternalProgressValue(aCurrAudioPlayProgressValue);
    
    this->refreshTextlineImageData(aProgressValue);
    
    VisualGraphics *theVisualGraphics = VisualGraphics::getInstance();
    
    VisualCamera aCamera;
    aCamera.setOrthographicProjection();
    
    size_t aMaxPixelWidth = theVisualGraphics->getCanvasPixelWidth();
    size_t aLeftStartPixelPos = (size_t)((aMaxPixelWidth - this->longestTextLinePixelWidth) / 2.0);
    
    BottomLeftPositionedPixelRect aClippingRect;
    aClippingRect.pixelRect = this->getViewRect();
    aClippingRect.bottomLeftPixel.x = aLeftStartPixelPos;
    size_t aVerticalPixelOffset = theVisualGraphics->getCanvasPixelHeight() - aClippingRect.pixelRect.height;
    aClippingRect.bottomLeftPixel.y = aVerticalPixelOffset;
    theVisualGraphics->setClippingRect(aClippingRect);
    
    double anAdditionalCoordOffset = 0.0;
    if (this->completeTextLinesPixelHeight() > this->displayedPixelHeight)
    {
        // we need to scroll vertically
        uint32 aPixelHeightDelta = this->completeTextLinesPixelHeight() - this->displayedPixelHeight;
        std::vector<VisualTextAutoScrollViewTextLineImageRecord *>::iterator aCurrFirstImageIterator = textScrollViewImageVector.begin() + this->firstTextLineIndexToShow;
        size_t aCurrDeltaPixelPosition = (*aCurrFirstImageIterator)->textImageVerticalPixelPosition;
        float aCurrPixelProgressValue = ((float)aCurrDeltaPixelPosition / (float)aPixelHeightDelta) * (float)aPixelHeightDelta;
        float aCurrPlayVerticalPixelPosition = aProgressValue * (float)aPixelHeightDelta;
        
        float aPixelDeficit = aCurrPlayVerticalPixelPosition - aCurrPixelProgressValue;
        anAdditionalCoordOffset = theVisualGraphics->yPixelToCoord(aPixelDeficit, aCamera);
    }
    
    double aMinStartCoordX = aCamera.getMaxLeftCoord() + theVisualGraphics->xPixelToCoord(aLeftStartPixelPos, aCamera);
    double aStartCoordX = aMinStartCoordX;

    VisualStringStyle& aStringStyle = this->textScrollViewStyledString->getStyle();
    RGBAColor aTextColor = aStringStyle.fontColor;
    size_t anAccumulatedHeight = 0;
    for (size_t i = this->firstTextLineIndexToShow; i <= this->lastTextLineIndexToShow; i++)
    {
        std::vector<VisualTextAutoScrollViewTextLineImageRecord *>::iterator anIterator = textScrollViewImageVector.begin() + i;
        
        VisualImage *aCurrImage = (*anIterator)->textImage;
        if (aCurrImage->isEmpty() == false)
        {
            float aScaleFactor = 1.0;
            
            if (aStringStyle.horizontalAlignment == kCenterAligned)
            {
                size_t aHorizontalAlignmentDelta = (size_t)((this->longestTextLinePixelWidth - (*anIterator)->textImagePixelRect.width) / 2.0);
                aStartCoordX = aMinStartCoordX + theVisualGraphics->xPixelToCoord(aHorizontalAlignmentDelta, aCamera);
            }
            else if (aStringStyle.horizontalAlignment == kRightAligned)
            {
                size_t aHorizontalAlignmentDelta = this->longestTextLinePixelWidth - (*anIterator)->textImagePixelRect.width;
                aStartCoordX = aMinStartCoordX + theVisualGraphics->xPixelToCoord(aHorizontalAlignmentDelta, aCamera);
            }
            
            Coord aCoord;
            TexCoord aTexCoord;
            VertexChain aVertexChain;
            
            // top left
            aCoord.x = aStartCoordX;
            aCoord.y = (aCamera.getMaxTopCoord() - theVisualGraphics->yPixelToCoord(anAccumulatedHeight, aCamera)) * aScaleFactor;
            aCoord.y += anAdditionalCoordOffset;
            aCoord.z = aCamera.getMaxNearPos();
            aTexCoord.s = 0.0;
            aTexCoord.t = aCurrImage->getLogicalHeight();
            VisualVertex* aTopLeftFrontVertex = new VisualVertex(aCoord, aTexCoord, aTextColor);
            aVertexChain.push_back(aTopLeftFrontVertex);
            
            // bottom left
            aCoord.x = aStartCoordX;
            aCoord.y = (aCamera.getMaxTopCoord() - theVisualGraphics->yPixelToCoord(anAccumulatedHeight + aCurrImage->getHeight(), aCamera)) * aScaleFactor;
            aCoord.y += anAdditionalCoordOffset;
            aCoord.z = aCamera.getMaxNearPos();
            aTexCoord.s = 0.0;
            aTexCoord.t = 0.0;
            VisualVertex* aBottomLeftFrontVertex = new VisualVertex(aCoord, aTexCoord, aTextColor);
            aVertexChain.push_back(aBottomLeftFrontVertex);
            
            // bottom right
            aCoord.x = (aStartCoordX + theVisualGraphics->xPixelToCoord(aCurrImage->getWidth(), aCamera)) * aScaleFactor;
            aCoord.y = (aCamera.getMaxTopCoord() - theVisualGraphics->yPixelToCoord(anAccumulatedHeight + aCurrImage->getHeight(), aCamera)) * aScaleFactor;
            aCoord.y += anAdditionalCoordOffset;
            aCoord.z = aCamera.getMaxNearPos();
            aTexCoord.s = aCurrImage->getLogicalWidth();
            aTexCoord.t = 0.0;
            VisualVertex* aBottomRightFrontVertex = new VisualVertex(aCoord, aTexCoord, aTextColor);
            aVertexChain.push_back(aBottomRightFrontVertex);
            
            // top right
            aCoord.x = (aStartCoordX + theVisualGraphics->xPixelToCoord(aCurrImage->getWidth(), aCamera)) * aScaleFactor;
            aCoord.y = (aCamera.getMaxTopCoord() - theVisualGraphics->yPixelToCoord(anAccumulatedHeight, aCamera)) * aScaleFactor;
            aCoord.y += anAdditionalCoordOffset;
            aCoord.z = aCamera.getMaxNearPos();
            aTexCoord.s = aCurrImage->getLogicalWidth();
            aTexCoord.t = aCurrImage->getLogicalHeight();
            VisualVertex* aTopRightFrontVertex = new VisualVertex(aCoord, aTexCoord, aTextColor);
            aVertexChain.push_back(aTopRightFrontVertex);
            
            aCurrImage->draw(const_cast<VertexChain*>(&aVertexChain));
            
            std::vector<VisualVertex *>::iterator it;
            for (it = aVertexChain.begin(); it != aVertexChain.end(); it++)
            {
                delete *it;
            }
        }
        
        anAccumulatedHeight += (*anIterator)->textImagePixelRect.height;
	}
    
    theVisualGraphics->disableClipping();
}


size_t VisualTextAutoScrollView::completeTextLinesPixelHeight()
{
    if (this->textScrollViewImageVector.empty())
    {
        return 0;
    }
    VisualTextAutoScrollViewTextLineImageRecord* aLastTextScrollViewTextLineImageRecord = this->textScrollViewImageVector.back();
    size_t aCompleteTextLinesPixelHeight = aLastTextScrollViewTextLineImageRecord->textImageVerticalPixelPosition + aLastTextScrollViewTextLineImageRecord->textImagePixelRect.height;
    return aCompleteTextLinesPixelHeight;
}


float VisualTextAutoScrollView::convertProgressValueToInternalProgressValue(float aCurrAudioPlayProgressValue)
{
    // return 0.0 for the first 5% of the play duration
    // return 1.0 for the last 5% of the play duration
    
    float aStretchDelta = 0.0;
    if (aCurrAudioPlayProgressValue < 0.5)
    {
        aStretchDelta = (0.5f - aCurrAudioPlayProgressValue) * 2.0f;
        aStretchDelta *= 0.05f;
        aCurrAudioPlayProgressValue -= aStretchDelta;
    }
    else
    {
        aStretchDelta = (aCurrAudioPlayProgressValue - 0.5f) * 2.0f;
        aStretchDelta *= 0.05f;
        aCurrAudioPlayProgressValue += aStretchDelta;
    }
    if (aCurrAudioPlayProgressValue > 1.0)
    {
        return 1.0;
    }
    else if (aCurrAudioPlayProgressValue < 0.0)
    {
        return 0.0;
    }
    
    // ease in and ease out
    
    float aProgressValue = (-1.0f * cos(aCurrAudioPlayProgressValue * (float)M_PI) * 0.5f) + 0.5f;
    return aProgressValue;
}


PixelRect VisualTextAutoScrollView::getViewRect(void)
{
    PixelRect aPixelRect;
    aPixelRect.width = 0;
    aPixelRect.height = 0;
    if (this->textScrollViewStyledString != NULL)
    {
        size_t aPixelHeight = this->displayedPixelHeight;
        aPixelRect.width = this->longestTextLinePixelWidth;
        aPixelRect.height = (this->completeTextLinesPixelHeight() < aPixelHeight) ? this->completeTextLinesPixelHeight() : aPixelHeight;
    }
    return aPixelRect;
}


void VisualTextAutoScrollView::setDisplayedPixelHeight(size_t aDisplayedPixelHeight)
{
    this->displayedPixelHeight = aDisplayedPixelHeight;
}
